#ifndef TocParserTokens_h
#define TocParserTokens_h
/* TocParserTokens.h -- List of labelled tokens and stuff
 *
 * Generated from: ./TocParser.g
 *
 * Terence Parr, Will Cohen, and Hank Dietz: 1989-2001
 * Purdue University Electrical Engineering
 * ANTLR Version 1.33MR32
 */
enum ANTLRTokenType {
	Eof=1,
	Comment=3,
	BeginString=5,
	Integer=6,
	TrackDef=7,
	FirstTrackNo=8,
	Audio=9,
	Mode0=10,
	Mode1=11,
	Mode1Raw=12,
	Mode2=13,
	Mode2Raw=14,
	Mode2Form1=15,
	Mode2Form2=16,
	Mode2FormMix=17,
	SubChanRw=18,
	SubChanRwRaw=19,
	Index=20,
	Catalog=21,
	Isrc=22,
	No=23,
	Copy=24,
	PreEmphasis=25,
	TwoChannel=26,
	FourChannel=27,
	DataFile=31,
	Fifo=32,
	Silence=33,
	Zero=34,
	Pregap=35,
	Start=36,
	End=37,
	TocTypeCdda=38,
	TocTypeCdrom=39,
	TocTypeCdromXa=40,
	TocTypeCdi=41,
	Swap=42,
	CdText=43,
	Language=44,
	LanguageMap=45,
	Title=46,
	Performer=47,
	Songwriter=48,
	Composer=49,
	Arranger=50,
	Message=51,
	DiscId=52,
	Genre=53,
	TocInfo1=54,
	TocInfo2=55,
	Reserved1=56,
	Reserved2=57,
	Reserved3=58,
	Reserved4=59,
	Closed=60,
	UpcEan=61,
	SizeInfo=62,
	LangEn=63,
	EncodingLatin=64,
	EncodingAscii=65,
	EncodingJis=66,
	EncodingKorean=67,
	EncodingMandarin=68,
	EndString=69,
	StringQuote=70,
	StringBackslash=71,
	StringOctal=72,
	String=73,
	DLGminToken=0,
	DLGmaxToken=9999};

#endif
