#ifndef TocLexerBase_h
#define TocLexerBase_h
/*
 * D L G L e x e r  C l a s s  D e f i n i t i o n
 *
 * Generated from: TocLexer.dlg
 *
 * 1989-2001 by  Will Cohen, Terence Parr, and Hank Dietz
 * Purdue University Electrical Engineering
 * DLG Version 1.33MR32
 */


#include "DLexerBase.h"

class TocLexerBase : public DLGLexerBase {
public:
public:
	static const int MAX_MODE;
	static const int DfaStates;
	static const int START;
	static const int STRING;
	typedef unsigned short DfaState;

	TocLexerBase(DLGInputStream *in,
		unsigned bufsize=2000)
		: DLGLexerBase(in, bufsize, 0)
	{
	;
	}
	void	  mode(int);
	ANTLRTokenType nextTokenType(void);
	void     advance(void);
protected:
	ANTLRTokenType act1();
	ANTLRTokenType act2();
	ANTLRTokenType act3();
	ANTLRTokenType act4();
	ANTLRTokenType act5();
	ANTLRTokenType act6();
	ANTLRTokenType act7();
	ANTLRTokenType act8();
	ANTLRTokenType act9();
	ANTLRTokenType act10();
	ANTLRTokenType act11();
	ANTLRTokenType act12();
	ANTLRTokenType act13();
	ANTLRTokenType act14();
	ANTLRTokenType act15();
	ANTLRTokenType act16();
	ANTLRTokenType act17();
	ANTLRTokenType act18();
	ANTLRTokenType act19();
	ANTLRTokenType act20();
	ANTLRTokenType act21();
	ANTLRTokenType act22();
	ANTLRTokenType act23();
	ANTLRTokenType act24();
	ANTLRTokenType act25();
	ANTLRTokenType act26();
	ANTLRTokenType act27();
	ANTLRTokenType act28();
	ANTLRTokenType act29();
	ANTLRTokenType act30();
	ANTLRTokenType act31();
	ANTLRTokenType act32();
	ANTLRTokenType act33();
	ANTLRTokenType act34();
	ANTLRTokenType act35();
	ANTLRTokenType act36();
	ANTLRTokenType act37();
	ANTLRTokenType act38();
	ANTLRTokenType act39();
	ANTLRTokenType act40();
	ANTLRTokenType act41();
	ANTLRTokenType act42();
	ANTLRTokenType act43();
	ANTLRTokenType act44();
	ANTLRTokenType act45();
	ANTLRTokenType act46();
	ANTLRTokenType act47();
	ANTLRTokenType act48();
	ANTLRTokenType act49();
	ANTLRTokenType act50();
	ANTLRTokenType act51();
	ANTLRTokenType act52();
	ANTLRTokenType act53();
	ANTLRTokenType act54();
	ANTLRTokenType act55();
	ANTLRTokenType act56();
	ANTLRTokenType act57();
	ANTLRTokenType act58();
	ANTLRTokenType act59();
	ANTLRTokenType act60();
	ANTLRTokenType act61();
	ANTLRTokenType act62();
	ANTLRTokenType act63();
	ANTLRTokenType act64();
	ANTLRTokenType act65();
	ANTLRTokenType act66();
	ANTLRTokenType act67();
	ANTLRTokenType act68();
	ANTLRTokenType act69();
	ANTLRTokenType act70();
	ANTLRTokenType act71();
	ANTLRTokenType act72();
	ANTLRTokenType act73();
	ANTLRTokenType act74();
	ANTLRTokenType act75();
	ANTLRTokenType act76();
	ANTLRTokenType act77();
	ANTLRTokenType act78();
	ANTLRTokenType act79();
	static DfaState st0[46];
	static DfaState st1[46];
	static DfaState st2[46];
	static DfaState st3[46];
	static DfaState st4[46];
	static DfaState st5[46];
	static DfaState st6[46];
	static DfaState st7[46];
	static DfaState st8[46];
	static DfaState st9[46];
	static DfaState st10[46];
	static DfaState st11[46];
	static DfaState st12[46];
	static DfaState st13[46];
	static DfaState st14[46];
	static DfaState st15[46];
	static DfaState st16[46];
	static DfaState st17[46];
	static DfaState st18[46];
	static DfaState st19[46];
	static DfaState st20[46];
	static DfaState st21[46];
	static DfaState st22[46];
	static DfaState st23[46];
	static DfaState st24[46];
	static DfaState st25[46];
	static DfaState st26[46];
	static DfaState st27[46];
	static DfaState st28[46];
	static DfaState st29[46];
	static DfaState st30[46];
	static DfaState st31[46];
	static DfaState st32[46];
	static DfaState st33[46];
	static DfaState st34[46];
	static DfaState st35[46];
	static DfaState st36[46];
	static DfaState st37[46];
	static DfaState st38[46];
	static DfaState st39[46];
	static DfaState st40[46];
	static DfaState st41[46];
	static DfaState st42[46];
	static DfaState st43[46];
	static DfaState st44[46];
	static DfaState st45[46];
	static DfaState st46[46];
	static DfaState st47[46];
	static DfaState st48[46];
	static DfaState st49[46];
	static DfaState st50[46];
	static DfaState st51[46];
	static DfaState st52[46];
	static DfaState st53[46];
	static DfaState st54[46];
	static DfaState st55[46];
	static DfaState st56[46];
	static DfaState st57[46];
	static DfaState st58[46];
	static DfaState st59[46];
	static DfaState st60[46];
	static DfaState st61[46];
	static DfaState st62[46];
	static DfaState st63[46];
	static DfaState st64[46];
	static DfaState st65[46];
	static DfaState st66[46];
	static DfaState st67[46];
	static DfaState st68[46];
	static DfaState st69[46];
	static DfaState st70[46];
	static DfaState st71[46];
	static DfaState st72[46];
	static DfaState st73[46];
	static DfaState st74[46];
	static DfaState st75[46];
	static DfaState st76[46];
	static DfaState st77[46];
	static DfaState st78[46];
	static DfaState st79[46];
	static DfaState st80[46];
	static DfaState st81[46];
	static DfaState st82[46];
	static DfaState st83[46];
	static DfaState st84[46];
	static DfaState st85[46];
	static DfaState st86[46];
	static DfaState st87[46];
	static DfaState st88[46];
	static DfaState st89[46];
	static DfaState st90[46];
	static DfaState st91[46];
	static DfaState st92[46];
	static DfaState st93[46];
	static DfaState st94[46];
	static DfaState st95[46];
	static DfaState st96[46];
	static DfaState st97[46];
	static DfaState st98[46];
	static DfaState st99[46];
	static DfaState st100[46];
	static DfaState st101[46];
	static DfaState st102[46];
	static DfaState st103[46];
	static DfaState st104[46];
	static DfaState st105[46];
	static DfaState st106[46];
	static DfaState st107[46];
	static DfaState st108[46];
	static DfaState st109[46];
	static DfaState st110[46];
	static DfaState st111[46];
	static DfaState st112[46];
	static DfaState st113[46];
	static DfaState st114[46];
	static DfaState st115[46];
	static DfaState st116[46];
	static DfaState st117[46];
	static DfaState st118[46];
	static DfaState st119[46];
	static DfaState st120[46];
	static DfaState st121[46];
	static DfaState st122[46];
	static DfaState st123[46];
	static DfaState st124[46];
	static DfaState st125[46];
	static DfaState st126[46];
	static DfaState st127[46];
	static DfaState st128[46];
	static DfaState st129[46];
	static DfaState st130[46];
	static DfaState st131[46];
	static DfaState st132[46];
	static DfaState st133[46];
	static DfaState st134[46];
	static DfaState st135[46];
	static DfaState st136[46];
	static DfaState st137[46];
	static DfaState st138[46];
	static DfaState st139[46];
	static DfaState st140[46];
	static DfaState st141[46];
	static DfaState st142[46];
	static DfaState st143[46];
	static DfaState st144[46];
	static DfaState st145[46];
	static DfaState st146[46];
	static DfaState st147[46];
	static DfaState st148[46];
	static DfaState st149[46];
	static DfaState st150[46];
	static DfaState st151[46];
	static DfaState st152[46];
	static DfaState st153[46];
	static DfaState st154[46];
	static DfaState st155[46];
	static DfaState st156[46];
	static DfaState st157[46];
	static DfaState st158[46];
	static DfaState st159[46];
	static DfaState st160[46];
	static DfaState st161[46];
	static DfaState st162[46];
	static DfaState st163[46];
	static DfaState st164[46];
	static DfaState st165[46];
	static DfaState st166[46];
	static DfaState st167[46];
	static DfaState st168[46];
	static DfaState st169[46];
	static DfaState st170[46];
	static DfaState st171[46];
	static DfaState st172[46];
	static DfaState st173[46];
	static DfaState st174[46];
	static DfaState st175[46];
	static DfaState st176[46];
	static DfaState st177[46];
	static DfaState st178[46];
	static DfaState st179[46];
	static DfaState st180[46];
	static DfaState st181[46];
	static DfaState st182[46];
	static DfaState st183[46];
	static DfaState st184[46];
	static DfaState st185[46];
	static DfaState st186[46];
	static DfaState st187[46];
	static DfaState st188[46];
	static DfaState st189[46];
	static DfaState st190[46];
	static DfaState st191[46];
	static DfaState st192[46];
	static DfaState st193[46];
	static DfaState st194[46];
	static DfaState st195[46];
	static DfaState st196[46];
	static DfaState st197[46];
	static DfaState st198[46];
	static DfaState st199[46];
	static DfaState st200[46];
	static DfaState st201[46];
	static DfaState st202[46];
	static DfaState st203[46];
	static DfaState st204[46];
	static DfaState st205[46];
	static DfaState st206[46];
	static DfaState st207[46];
	static DfaState st208[46];
	static DfaState st209[46];
	static DfaState st210[46];
	static DfaState st211[46];
	static DfaState st212[46];
	static DfaState st213[46];
	static DfaState st214[46];
	static DfaState st215[46];
	static DfaState st216[46];
	static DfaState st217[46];
	static DfaState st218[46];
	static DfaState st219[46];
	static DfaState st220[46];
	static DfaState st221[46];
	static DfaState st222[46];
	static DfaState st223[46];
	static DfaState st224[46];
	static DfaState st225[46];
	static DfaState st226[46];
	static DfaState st227[46];
	static DfaState st228[46];
	static DfaState st229[46];
	static DfaState st230[46];
	static DfaState st231[46];
	static DfaState st232[46];
	static DfaState st233[46];
	static DfaState st234[46];
	static DfaState st235[46];
	static DfaState st236[46];
	static DfaState st237[46];
	static DfaState st238[46];
	static DfaState st239[46];
	static DfaState st240[46];
	static DfaState st241[46];
	static DfaState st242[46];
	static DfaState st243[46];
	static DfaState st244[46];
	static DfaState st245[46];
	static DfaState st246[46];
	static DfaState st247[46];
	static DfaState st248[46];
	static DfaState st249[46];
	static DfaState st250[46];
	static DfaState st251[46];
	static DfaState st252[46];
	static DfaState st253[46];
	static DfaState st254[46];
	static DfaState st255[46];
	static DfaState st256[46];
	static DfaState st257[46];
	static DfaState st258[46];
	static DfaState st259[46];
	static DfaState st260[46];
	static DfaState st261[46];
	static DfaState st262[46];
	static DfaState st263[46];
	static DfaState st264[46];
	static DfaState st265[46];
	static DfaState st266[46];
	static DfaState st267[46];
	static DfaState st268[46];
	static DfaState st269[46];
	static DfaState st270[46];
	static DfaState st271[46];
	static DfaState st272[46];
	static DfaState st273[46];
	static DfaState st274[46];
	static DfaState st275[46];
	static DfaState st276[46];
	static DfaState st277[46];
	static DfaState st278[46];
	static DfaState st279[46];
	static DfaState st280[46];
	static DfaState st281[46];
	static DfaState st282[46];
	static DfaState st283[46];
	static DfaState st284[46];
	static DfaState st285[46];
	static DfaState st286[46];
	static DfaState st287[46];
	static DfaState st288[46];
	static DfaState st289[46];
	static DfaState st290[46];
	static DfaState st291[46];
	static DfaState st292[46];
	static DfaState st293[46];
	static DfaState st294[46];
	static DfaState st295[46];
	static DfaState st296[46];
	static DfaState st297[46];
	static DfaState st298[46];
	static DfaState st299[46];
	static DfaState st300[46];
	static DfaState st301[46];
	static DfaState st302[46];
	static DfaState st303[46];
	static DfaState st304[46];
	static DfaState st305[46];
	static DfaState st306[46];
	static DfaState st307[46];
	static DfaState st308[46];
	static DfaState st309[46];
	static DfaState st310[46];
	static DfaState st311[46];
	static DfaState st312[46];
	static DfaState st313[46];
	static DfaState st314[46];
	static DfaState st315[46];
	static DfaState st316[46];
	static DfaState st317[46];
	static DfaState st318[46];
	static DfaState st319[46];
	static DfaState st320[46];
	static DfaState st321[46];
	static DfaState st322[46];
	static DfaState st323[46];
	static DfaState st324[46];
	static DfaState st325[46];
	static DfaState st326[46];
	static DfaState st327[46];
	static DfaState st328[46];
	static DfaState st329[46];
	static DfaState st330[46];
	static DfaState st331[46];
	static DfaState st332[6];
	static DfaState st333[6];
	static DfaState st334[6];
	static DfaState st335[6];
	static DfaState st336[6];
	static DfaState st337[6];
	static DfaState st338[6];
	static DfaState st339[6];
	static DfaState st340[6];
	static DfaState st341[6];
	static DfaState *dfa[342];
	static DfaState dfa_base[];
	static unsigned char *b_class_no[];
	static DfaState accepts[343];
	static DLGChar alternatives[343];
	static ANTLRTokenType (TocLexerBase::*actions[80])();
	static unsigned char shift0[257];
	static unsigned char shift1[257];
	int ZZSHIFT(int c) { return b_class_no[automaton][1+c]; }
//
// 133MR1 Deprecated feature to allow inclusion of user-defined code in DLG class header
//
#ifdef DLGLexerIncludeFile
#include DLGLexerIncludeFile
#endif
};
typedef ANTLRTokenType (TocLexerBase::*PtrTocLexerBaseMemberFunc)();
#endif
